@@ -10,6 +10,7 @@
 
 #include <marpaWrapper.h>
 #include <genericStack.h>
+#include <genericHash.h>
 #include <genericLogger.h>
 #include <pcre2.h>
 #include <tconv.h>
@@ -159,10 +160,12 @@ struct marpaESLIFSymbol {
   int                          priorityi;              /* Symbol priority */
   char                        *actions;                /* Action */
   unsigned int                 nbupdatei;              /* Number of updates - used in grammar ESLIF actions */
-  genericStack_t              *nullableRuleStackp;     /* Used during validation, to determine nullable semantics */
+  genericStack_t               _nullableRuleStack;     /* Used during validation, to determine nullable semantics */
+  genericStack_t              *nullableRuleStackp;     /* Pointer to _nullableRuleStack */
   char                        *nullableActions;        /* Nullable semantic */
   int                          propertyBitSet;
-  genericStack_t              *lhsRuleStackp;          /* Stack of rules having this symbol as LHS */
+  genericStack_t               _lhsRuleStack;          /* Stack of rules having this symbol as LHS */
+  genericStack_t              *lhsRuleStackp;          /* Pointer to stack of rules having this symbol as LHS */
   short                        haveExceptionb;         /* If true, this is the symbol before the '-' character of an exception */
   marpaESLIF_symbol_t         *exceptionp;             /* Pointer to the exception itself, the one after the '-' character */
 };
@@ -175,7 +178,8 @@ struct marpaESLIF_rule {
   char                  *asciishows;                   /* Rule show (ASCII) */
   marpaESLIF_symbol_t   *lhsp;                         /* LHS symbol */
   marpaESLIF_symbol_t   *separatorp;                   /* Eventual separator symbol */
-  genericStack_t        *rhsStackp;                    /* Stack of RHS symbols */
+  genericStack_t         _rhsStack;                    /* Stack of RHS symbols */
+  genericStack_t        *rhsStackp;                    /* Pointer to stack of RHS symbols */
   marpaESLIF_symbol_t   *exceptionp;                   /* Exception symbol */
   int                    exceptionIdi;                 /* Exception symbol Id */
   char                  *actions;                      /* Action */
@@ -192,7 +196,7 @@ struct marpaESLIF_rule {
 
 /* A grammar */
 struct marpaESLIF_grammar {
-  marpaESLIF_t          *marpaESLIFp;                        /* Shallow pointer to top marpaESLIFp */
+  marpaESLIFGrammar_t   *marpaESLIFGrammarp;                 /* Shallow pointer to parent structure marpaESLIFGrammarp */
   marpaESLIF_grammar_t  *selfp;                              /* This is vicious: internally the grammar is duplicated on the stack. This pointer give the original allocated memory. */
   int                    leveli;                             /* Grammar level */
   marpaESLIF_string_t   *descp;                              /* Grammar description */
@@ -203,8 +207,10 @@ struct marpaESLIF_grammar {
   marpaWrapperGrammar_t *marpaWrapperGrammarDiscardp;        /* Grammar implementation at :discard */
   marpaWrapperGrammar_t *marpaWrapperGrammarDiscardNoEventp; /* Grammar implementation at :discard forcing no event */
   marpaESLIF_symbol_t   *discardp;                           /* Discard symbol, used at grammar validation */
-  genericStack_t        *symbolStackp;                       /* Stack of symbols */
-  genericStack_t        *ruleStackp;                         /* Stack of rules */
+  genericStack_t         _symbolStack;                       /* Stack of symbols */
+  genericStack_t        *symbolStackp;                       /* Pointer to stack of symbols */
+  genericStack_t         _ruleStack;                         /* Stack of rules */
+  genericStack_t        *ruleStackp;                         /* Pointer to stack of rules */
   char                  *defaultSymbolActions;               /* Default action for symbols */
   char                  *defaultRuleActions;                 /* Default action for rules */
   char                  *defaultFreeActions;                 /* Default action for free */
@@ -230,11 +236,13 @@ struct marpaESLIF {
   marpaESLIF_terminal_t *characterClassModifiersp;    /* Internal regex for match character class modifiers */
   marpaESLIF_terminal_t *regexModifiersp;             /* Internal regex for match regex modifiers */
   genericLogger_t       *traceLoggerp;                /* For cases where this is silent mode but compiled with TRACE */
+  short                  NULLisZeroBytesb;            /* An internal boolean to help when we can safely do calloc() */
 };
 
 struct marpaESLIFGrammar {
   marpaESLIF_t             *marpaESLIFp;
-  genericStack_t           *grammarStackp;
+  genericStack_t            _grammarStack;     /* Stack of grammars */
+  genericStack_t           *grammarStackp;     /* Pointer to stack of grammars */
   marpaESLIF_grammar_t     *grammarp;          /* This is a SHALLOW copy of current grammar in grammarStackp, defaulting to the top grammar */
   short                     warningIsErrorb;   /* Current warningIsErrorb setting (used when parsing grammars ) */
   short                     warningIsIgnoredb; /* Current warningIsErrorb setting (used when parsing grammars ) */
@@ -261,11 +269,18 @@ struct marpaESLIFRecognizer {
   /* except in very precise situations (typically the new()/free() or when */
   /* faking a new() method). */
   marpaESLIF_t                *marpaESLIFp;
+
+  /* Because recognizers can be cached we cannot afford marpaESLIFGrammarp to point to something on the stack when we free: Grammarp and grammarp */
+  /* We guarantee that pointers below grammarp are correct (these are shallow pointers) */
+  marpaESLIFGrammar_t          _marpaESLIFGrammar;
+  marpaESLIF_grammar_t         _grammar;
   marpaESLIFGrammar_t         *marpaESLIFGrammarp;
+  
   marpaESLIFRecognizerOption_t marpaESLIFRecognizerOption;
   marpaWrapperRecognizer_t    *marpaWrapperRecognizerp; /* Current recognizer */
   marpaWrapperGrammar_t       *marpaWrapperGrammarp; /* Shallow copy of cached grammar in use */
-  genericStack_t              *lexemeInputStackp;  /* Internal input stack of lexemes */
+  genericStack_t               _lexemeInputStack;  /* Internal input stack of lexemes */
+  genericStack_t              *lexemeInputStackp;  /* Pointer to internal input stack of lexemes */
   marpaESLIFEvent_t           *eventArrayp;        /* For the events */
   size_t                       eventArrayl;        /* Current number of events */
   size_t                       eventArraySizel;    /* Real allocated size (to avoid constant free/deletes) */
@@ -277,6 +292,7 @@ struct marpaESLIFRecognizer {
   int                          resumeCounteri;    /* Internal counter for tracing - no functional impact */
   int                          callstackCounteri; /* Internal counter for tracing - no functional impact */
 
+  /* ------------------ Internal elements that are shared with all children ------------------------ */
   char                        *_buffers;       /* Pointer to allocated buffer containing input */
   size_t                       _bufferl;       /* Number of valid bytes in this buffer (!= allocated size) */
   size_t                       _bufferallocl;  /* Number of allocated bytes in this buffer (!= valid bytes) */
@@ -284,7 +300,8 @@ struct marpaESLIFRecognizer {
   short                        _eofb;          /* EOF flag */
   short                        _utfb;          /* A flag to say if input is UTF-8 correct. Automatically true if _charconv is true. Can set be regex engine as well. */
   short                        _charconvb;     /* A flag to say if latest stream chunk was converted to UTF-8 */
-  marpaWrapperGrammar_t     ***_marpaWrapperGrammarCacheppp; /* Cache of all needed cloned()/precomputed grammars */
+  genericHash_t                _marpaESLIFRecognizerHash; /* Cache of recognizers ready for re-use */
+  /* --------------- End of internal elements that are shared with all children --------------------- */
 
   int                          leveli;         /* Recognizer level (!= grammar level) */
 
@@ -295,7 +312,6 @@ struct marpaESLIFRecognizer {
   short                       *eofbp;          /* Ditto for the EOF flag */
   short                       *utfbp;          /* Ditto for the UTF-8 correctness flag */
   short                       *charconvbp;     /* Ditto for the character conversion flag */
-  marpaWrapperGrammar_t    ****marpaWrapperGrammarCachepppp; /* Ditto for grammars cache */
 
   size_t                       parentDeltal;   /* Parent original delta - used to recovert parent current pointer at our free */
   char                        *inputs;         /* Current pointer in input - specific to every recognizer */
@@ -326,9 +342,12 @@ struct marpaESLIFRecognizer {
   short                        exhaustedb;     /* Internally, every recognizer need to know if parsing is exhausted */
   short                        completedb;     /* Ditto for completion (used in case od discard events) */
   short                        continueb;
-  genericStack_t              *alternativeStackSymbolp;          /* Current alternative stack containing symbol information and the matched size */
-  genericStack_t              *commitedAlternativeStackSymbolp;  /* Commited alternative stack */
-  genericStack_t              *set2InputStackp;
+  genericStack_t               _alternativeStackSymbol;          /* Current alternative stack containing symbol information and the matched size */
+  genericStack_t              *alternativeStackSymbolp;          /* Pointer to current alternative stack containing symbol information and the matched size */
+  genericStack_t               _commitedAlternativeStackSymbol;  /* Commited alternative stack */
+  genericStack_t              *commitedAlternativeStackSymbolp;  /* Pointer to commited alternative stack */
+  genericStack_t               _set2InputStack;                  /* Mapping latest Earley Set to absolute input offset and length */
+  genericStack_t              *set2InputStackp;                  /* Pointer to mapping latest Earley Set to absolute input offset and length */
   char                       **lexemesArrayp;      /* Persistent buffer of last call to marpaESLIFRecognizer_lexeme_expectedb */
   size_t                       lexemesArrayAllocl; /* Current allocated size -; */
   short                       *discardEventStatebp; /* Discard current event states for the CURRENT grammar (marpaESLIFRecognizerp->marpaESLIFGrammarp->grammarp) */
@@ -338,6 +357,9 @@ struct marpaESLIFRecognizer {
   marpaESLIF_lexeme_data_t   **lastTrypp;           /* Lexeme or :discard last try for the CURRENT grammar */
   size_t                       minExceptionSizel;   /* Minimum total matched size before a completion event on a symbol to be checked for exception is considered */
   size_t                       totalMatchedSizel;   /* Total matched - used only when exceptionb is set */
+  short                        discardOnOffb;       /* Discard is on or off ? */
+  short                        pristineb;           /* 1: pristine, i.e. can be reused, 0: have at least one thing that happened at the raw grammar level, modulo the eventual initial events */
+  genericHash_t               *marpaESLIFRecognizerHashp; /* Ditto for recognizers cache */
 };
 
 struct marpaESLIF_lexeme_data {
